#
# Copyright 2018 Fairphone B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""The execution of system loads on devices."""

import concurrent
from concurrent.futures import ThreadPoolExecutor
import enum
import logging
from pathlib import PosixPath
from typing import Any, List
from uuid import UUID, uuid4

import adb.usb_exceptions
from android_tools import App, Device

from .loads import SystemLoad


LOG = logging.getLogger(__name__)


class DeviceRunner:
    """A runner to execute a session on a device.

    Runners of a session should be executed with the :py:func:`run` method in a
    parallel fashion (e.g. threads).

    Arguments:
        device: The device to run the session on.
        session: The session to run on the device.
    """

    def __init__(self, device: Device, session: "Session"):
        """Initialise the runner."""
        super().__init__()
        self.device = device
        self.session = session
        self._absolute_dir = PosixPath(device.storage_dir) / session.working_dir

    def __repr__(self) -> str:
        """Return an unambiguous representation of the runner."""
        return f"DeviceRunner<{self.device}/{self.session}>"

    def run(self) -> None:
        """Execute a session on the device.

        First prepare the device with the :py:func:`_prepare()` method. Then,
        starts an ADB shell to execute the session and drop it after a second.
        The load will run on the device as an orphaned process.

        Raises:
            ValueError: When the on-device command failed initialisation.

        """
        output = None
        run_cmd = self.prepare()

        self.device.unlock()
        self.device.home()

        LOG.debug("Executing command on %s: %s", self.device, run_cmd)
        with self.device.adb() as _adb:
            try:
                # HACK Do not keep the ADB communication alive.
                # Launch the command and trigger a timeout shortly after. This
                # will result in both this execution thread and the on-device
                # process to continue independently. Of course there is no way
                # to reattach to the on-device process and monitor it.
                # We do this that way because tests have proven that we cannot
                # rely on ADB to properly close the communication even when we
                # start a detached command (`command &`).
                output = _adb.Shell(run_cmd, timeout_ms=1000)
            except adb.usb_exceptions.ReadFailedError:
                # This is the expected timeout of the command.
                pass

        # The shell should not spit any error message, unless there was an error
        # initiating the command.
        if output:
            raise ValueError(f"Unexpected ADB shell output: {output}")

    def prepare(self) -> str:
        """Prepare the device to execute the session.

        1. Create a working directory where assets will be pushed to and where
            log files will be gathered.
        2. Deploy the apps and other executable files required by the load.

        Returns:
            The on-device command to execute the session.

        """
        with self.device.adb() as _adb:
            _adb.Shell(f"mkdir -p {str(self._absolute_dir)}")

            for asset in self.session.load.assets:
                if isinstance(asset, App):
                    LOG.debug("Installing app %s on %s", asset, self.device)
                    self.device.install_app(asset)
                else:
                    LOG.error(
                        "Ignoring asset `%s`, unknown type: %s",
                        asset,
                        type(asset),
                    )

        return (
            f"cd {str(self._absolute_dir)} && "
            f"{self.session.load.build_command()}"
        )


class SessionState(enum.Enum):
    """A session state."""

    #: The session is created and waiting to be run.
    IDLE = "Idle"
    #: Running phase, the session is being started on the devices.
    STARTING = "Starting"
    #: Final running phase, the session has been started on the devices.
    STARTED = "Started"


class SessionDeviceState(enum.Enum):
    """A device state, related to a specific session."""

    #: The session has not been started on the device.
    IDLE = "Idle"
    #: The session is running on the device.
    RUNNING = "Running"
    #: The session has completed on the device.
    DONE = "Done"


class SessionError(Exception):
    """A session error.

    Arguments:
        session: The session that failed.
        reason: The reason of the failure.

    Attributes:
        reason: The reason of the failure.
        session: The session that failed.

    """

    def __init__(self, session: "Session", reason: str):
        """Initialise the exception."""
        super().__init__(f"Session {session.uuid} failed: {reason}")
        self.reason = reason
        self.session = session


class Session:
    """A reproducible unit made of a system load.

    Arguments:
        load: The system load.
        devices: The devices to run the session on.
        uuid (optional): The session :py:class:`uuid.UUID` (or its string
            representation). Generate a new :py:class:`uuid.UUID` if none is
            given.

    Attributes:
        devices: The devices to run the session on.
        load: The system load (:py:class:`.loads.SystemLoad`).
        state: The session state (:py:class:`SessionState`). Mainly useful to
            monitor the scheduling of a session. Follow a session evolution
            through its devices state instead (see :py:func:`get_device_state`).
        runners: The dictionary of the session execution on the devices, mapping
            :py:class:`concurrent.futures.Future` to
            :py:class:`android_tools.Device`.
        uuid: The session UUID (:py:class:`uuid.UUID`).

    """

    def __init__(self, load: SystemLoad, devices: List, uuid: Any = None):
        """Initialise the session."""
        super().__init__()
        self.load = load
        self.devices = devices
        self.uuid = UUID(uuid) if uuid else uuid4()
        self.runners = {}
        self.state = SessionState.IDLE

    def __repr__(self) -> str:
        """Return an unambiguous representation of the session."""
        return f"Session<{self.uuid}>"

    @property
    def working_dir(self) -> PosixPath:
        """Return the (relative) on-device working directory."""
        return PosixPath("maestro") / "sessions" / str(self.uuid)

    def get_device_state(self, device: Device) -> SessionDeviceState:
        """Compute the state of this session on a device.

        Access a device and investigate its current state.

        Attributes:
            device: The device to compute the state of.
        Returns:
            The :py:class:`SessionDeviceState` with regards to this session.

        """
        with device.adb() as _adb:
            _, _, log_mtime = _adb.Stat(
                str(
                    PosixPath(device.storage_dir)
                    / self.working_dir
                    / self.load.log_file
                )
            )
            if not log_mtime:
                state = SessionDeviceState.IDLE
            elif _adb.Shell(f"ps | grep {self.load.cmd}"):
                state = SessionDeviceState.RUNNING
            else:
                state = SessionDeviceState.DONE
        return state

    def run(self) -> None:
        """Execute the session.

        Delegate the session execution on the devices to
        :py:class:`DeviceRunner` objects. Each runner runs in its own thread.
        This method waits for the runner threads completion before returning.

        Raises:
            SessionError: If the session has been started already, or if there
                is no devices to start the session on. The session state will be
                respectively different (and equal) to
                :py:const:`SessionState.IDLE`.

        """
        if self.state is not SessionState.IDLE:
            raise SessionError(self, "Already started")
        if not self.devices:
            raise SessionError(self, "No devices to work on")

        LOG.debug("Starting session %s on %s", self.uuid, self.devices)
        self.state = SessionState.STARTING

        device_runners = [
            DeviceRunner(device=device, session=self) for device in self.devices
        ]
        with ThreadPoolExecutor(max_workers=len(self.devices)) as executor:
            self.runners = {
                executor.submit(DeviceRunner.run, dr): dr
                for dr in device_runners
            }

            # Wait until all the runners completed
            for future in concurrent.futures.as_completed(self.runners):
                runner = self.runners[future]
                error = future.exception()
                if error:
                    LOG.error(
                        "Execution failed on %s with: %s", runner.device, error
                    )
                else:
                    LOG.info("Execution started on %s", runner.device)

        self.state = SessionState.STARTED
